﻿using System;
using System.ComponentModel;
using System.Runtime.InteropServices;
using System.Windows.Forms;
using HResult = System.Int32;

namespace RevSoft {

	#region COM Interface

	[InterfaceType(ComInterfaceType.InterfaceIsIUnknown)]
	[Guid("13b5c004-4377-4c94-a8f9-efc1fdaeb31c")]
	public interface IWabbitemu {
		[PreserveSig()]
		HResult ShowMessage();
		[PreserveSig()]
		HResult GetState(uint uSlot, out Z80State state);
		[PreserveSig()]
		HResult SetState(uint uSlot, Z80State state);
		[PreserveSig()]
		HResult KeyPress(uint uSlot, Keys vk);
		[PreserveSig()]
		HResult KeyRelease(uint uSlot, Keys vk);
		[PreserveSig()]
		HResult ReadMem(uint uSlot, ushort address, out byte value);
		[PreserveSig()]
		HResult WriteMem(uint uSlot, ushort address, byte value);
		[PreserveSig()]
		HResult DrawScreen(uint uSlot, IntPtr data);
	}

	#endregion

	#region Types

	/// <summary>
	/// Represents the state of a Z80 CPU.
	/// </summary>
	[StructLayout(LayoutKind.Explicit)]
	public class Z80State {

		#region Type Converters

		class UInt16HexConverter : TypeConverter {
			public override bool CanConvertFrom(ITypeDescriptorContext context, Type sourceType) {
				return sourceType == typeof(string) || base.CanConvertFrom(context, sourceType);
			}
			public override bool CanConvertTo(ITypeDescriptorContext context, Type destinationType) {
				return destinationType == typeof(string) || base.CanConvertTo(context, destinationType);
			}
			public override object ConvertFrom(ITypeDescriptorContext context, System.Globalization.CultureInfo culture, object value) {
				string s = value as string;
				if (s != null) {
					return (s.StartsWith("$")) ? Convert.ToUInt16(s.Substring(1), 16) : Convert.ToUInt16(s);
				}
				return base.ConvertFrom(context, culture, value);
			}
			public override object ConvertTo(ITypeDescriptorContext context, System.Globalization.CultureInfo culture, object value, Type destinationType) {
				return destinationType == typeof(string)
					? "$" + ((ushort)value).ToString("X4")
					: base.ConvertTo(context, culture, value, destinationType);
			}
		}

		class ByteHexConverter : TypeConverter {
			public override bool CanConvertFrom(ITypeDescriptorContext context, Type sourceType) {
				return sourceType == typeof(string) || base.CanConvertFrom(context, sourceType);
			}
			public override bool CanConvertTo(ITypeDescriptorContext context, Type destinationType) {
				return destinationType == typeof(string) || base.CanConvertTo(context, destinationType);
			}
			public override object ConvertFrom(ITypeDescriptorContext context, System.Globalization.CultureInfo culture, object value) {
				string s = value as string;
				if (s != null) {
					return (s.StartsWith("$")) ? Convert.ToByte(s.Substring(1), 16) : Convert.ToByte(s);
				}
				return base.ConvertFrom(context, culture, value);
			}
			public override object ConvertTo(ITypeDescriptorContext context, System.Globalization.CultureInfo culture, object value, Type destinationType) {
				return destinationType == typeof(string)
					? "$" + ((byte)value).ToString("X2")
					: base.ConvertTo(context, culture, value, destinationType);
			}
		}  

		#endregion

		#region Private Fields

		[FieldOffset(00)] byte f;

		[FieldOffset(00)] ushort af;
		[FieldOffset(02)] ushort bc;
		[FieldOffset(04)] ushort hl;
		[FieldOffset(06)] ushort de;

		[FieldOffset(08)] ushort afSh;
		[FieldOffset(10)] ushort bcSh;
		[FieldOffset(12)] ushort hlSh;
		[FieldOffset(14)] ushort deSh;

		[FieldOffset(16)] ushort ix;
		[FieldOffset(18)] ushort iy;
		[FieldOffset(20)] ushort pc;
		[FieldOffset(22)] ushort sp;

		[FieldOffset(24)] byte i;
		[FieldOffset(25)] byte r;
		[FieldOffset(26)] byte bus;
		[FieldOffset(27)] byte iff1;
		[FieldOffset(28)] byte iff2;
		[FieldOffset(29)] byte imode;
		
		[FieldOffset(30)] double seconds;

		#endregion

		#region Public Properties

		#region Flags

		/// <summary>
		/// Gets or sets the state of the carry flag.
		/// </summary>
		[Category("Flags"), DisplayName("C"), Description("Carry flag.")]
		public bool FlagC { get { return (this.f & (1 << 0)) != 0; } set { this.f = (byte)((this.f & ~(1 << 0)) | (value ? (1 << 0) : 0)); } }
		/// <summary>
		/// Gets or sets the state of the subtraction flag.
		/// </summary>
		[Category("Flags"), DisplayName("\tN"), Description("Subtraction flag.")]
		public bool FlagN { get { return (this.f & (1 << 1)) != 0; } set { this.f = (byte)((this.f & ~(1 << 1)) | (value ? (1 << 1) : 0)); } }
		/// <summary>
		/// Gets or sets the state of the parity/overflow flag.
		/// </summary>
		[Category("Flags"), DisplayName("\t\tP/V"), Description("Parity/Overflow flag.")]
		public bool FlagPV { get { return (this.f & (1 << 2)) != 0; } set { this.f = (byte)((this.f & ~(1 << 2)) | (value ? (1 << 2) : 0)); } }
		/// <summary>
		/// Gets or sets the state of the BCD half-carry flag.
		/// </summary>
		[Category("Flags"), DisplayName("\t\t\tH"), Description("BCD half-carry flag.")]
		public bool FlagH { get { return (this.f & (1 << 4)) != 0; } set { this.f = (byte)((this.f & ~(1 << 4)) | (value ? (1 << 4) : 0)); } }
		/// <summary>
		/// Gets or sets the state of the zero flag.
		/// </summary>
		[Category("Flags"), DisplayName("\t\t\t\tZ"), Description("Zero flag.")]
		public bool FlagZ { get { return (this.f & (1 << 6)) != 0; } set { this.f = (byte)((this.f & ~(1 << 6)) | (value ? (1 << 6) : 0)); } }
		/// <summary>
		/// Gets or sets the state of the sign flag.
		/// </summary>
		[Category("Flags"), DisplayName("\t\t\t\t\tS"), Description("Sign flag.")]
		public bool FlagS { get { return (this.f & (1 << 7)) != 0; } set { this.f = (byte)((this.f & ~(1 << 7)) | (value ? (1 << 7) : 0)); } }

		#endregion

		#region Register Pairs

		/// <summary>
		/// Gets or sets the value of the AF register pair.
		/// </summary>
		[DisplayName("AF"), Category("Register Pairs"), TypeConverter(typeof(UInt16HexConverter))]
		public ushort RegisterAF {
			get { return this.af; }
			set { this.af = value; }
		}

		/// <summary>
		/// Gets or sets the value of the BC register pair.
		/// </summary>
		[DisplayName("BC"), Category("Register Pairs"), TypeConverter(typeof(UInt16HexConverter))]
		public ushort RegisterBC {
			get { return this.bc; }
			set { this.bc = value; }
		}

		/// <summary>
		/// Gets or sets the value of the DE register pair.
		/// </summary>
		[DisplayName("DE"), Category("Register Pairs"), TypeConverter(typeof(UInt16HexConverter))]
		public ushort RegisterDE {
			get { return this.de; }
			set { this.de = value; }
		}

		/// <summary>
		/// Gets or sets the value of the HL register pair.
		/// </summary>
		[DisplayName("HL"), Category("Register Pairs"), TypeConverter(typeof(UInt16HexConverter))]
		public ushort RegisterHL {
			get { return this.hl; }
			set { this.hl = value; }
		}

		/// <summary>
		/// Gets or sets the value of the AF shadow register pair.
		/// </summary>
		[DisplayName("AF'"), Category("Shadow Register Pairs"), TypeConverter(typeof(UInt16HexConverter))]
		public ushort ShadowRegisterAF {
			get { return this.afSh; }
			set { this.afSh = value; }
		}

		/// <summary>
		/// Gets or sets the value of the BC shadow register pair.
		/// </summary>
		[DisplayName("BC'"), Category("Shadow Register Pairs"), TypeConverter(typeof(UInt16HexConverter))]
		public ushort ShadowRegisterBC {
			get { return this.bcSh; }
			set { this.bcSh = value; }
		}

		/// <summary>
		/// Gets or sets the value of the DE shadow register pair.
		/// </summary>
		[DisplayName("DE'"), Category("Shadow Register Pairs"), TypeConverter(typeof(UInt16HexConverter))]
		public ushort ShadowRegisterDE {
			get { return this.deSh; }
			set { this.deSh = value; }
		}

		/// <summary>
		/// Gets or sets the value of the HL shadow register pair.
		/// </summary>
		[DisplayName("HL'"), Category("Shadow Register Pairs"), TypeConverter(typeof(UInt16HexConverter))]
		public ushort ShadowRegisterHL {
			get { return this.hlSh; }
			set { this.hlSh = value; }
		}

		#endregion

		#region Special-Purpose Registers

		/// <summary>
		/// Gets or sets the value of the interrupt vector register.
		/// </summary>
		[DisplayName("\tI"), Category("Special-Purpose Registers"), TypeConverter(typeof(ByteHexConverter))]
		public byte RegisterI {
			get { return this.i; }
			set { this.i = value; }
		}

		/// <summary>
		/// Gets or sets the value of the memory refresh register.
		/// </summary>
		[DisplayName("\tR"), Category("Special-Purpose Registers"), TypeConverter(typeof(ByteHexConverter))]
		public byte RegisterR {
			get { return this.r; }
			set { this.r = value; }
		}

		/// <summary>
		/// Gets or sets the value of the IX index register.
		/// </summary>
		[DisplayName("IX"), Category("Special-Purpose Registers"), TypeConverter(typeof(UInt16HexConverter))]
		public ushort RegisterIX {
			get { return this.ix; }
			set { this.ix = value; }
		}

		/// <summary>
		/// Gets or sets the value of the IY index register.
		/// </summary>
		[DisplayName("IY"), Category("Special-Purpose Registers"), TypeConverter(typeof(UInt16HexConverter))]
		public ushort RegisterIY {
			get { return this.iy; }
			set { this.iy = value; }
		}

		/// <summary>
		/// Gets or sets the value of the stack pointer register.
		/// </summary>
		[DisplayName("SP"), Category("Special-Purpose Registers"), TypeConverter(typeof(UInt16HexConverter))]
		public ushort RegisterSP {
			get { return this.sp; }
			set { this.sp = value; }
		}

		/// <summary>
		/// Gets or sets the value of the program counter register.
		/// </summary>
		[DisplayName("PC"), Category("Special-Purpose Registers"), TypeConverter(typeof(UInt16HexConverter))]
		public ushort RegisterPC {
			get { return this.pc; }
			set { this.pc = value; }
		}

		#endregion

		#region Interrupt Control

		/// <summary>
		/// Gets or sets the state of the interrupt flip-flop 1 (IFF1).
		/// </summary>
		[DisplayName("IFF1"), Category("Interrupt Control")]
		public bool InterruptFlipFlop1 {
			get { return this.iff1 != 0; }
			set { this.iff1 = (byte)(value ? 1 : 0); }
		}

		/// <summary>
		/// Gets or sets the state of the interrupt flip-flop 2 (IFF2).
		/// </summary>
		[DisplayName("IFF2"), Category("Interrupt Control")]
		public bool InterruptFlipFlop2 {
			get { return this.iff2 != 0; }
			set { this.iff2 = (byte)(value ? 1 : 0); }
		}

		/// <summary>
		/// Gets or sets the interrupt mode.
		/// </summary>
		[DisplayName("Interrupt Mode"), Category("Interrupt Control")]
		public byte InterruptMode {
			get { return this.imode; }
			set { this.imode = value; }
		}

		#endregion

		#region Time

		/// <summary>
		/// Gets the amount of time that the calculator has been running.
		/// </summary>
		[DisplayName("Running Time")]
		public TimeSpan RunningTime {
			get { return TimeSpan.FromSeconds(this.seconds); }
		}

		#endregion

		#endregion
	}

	#endregion

	/// <summary>
	/// Provides methods to interact with an instance of Wabbitemu.
	/// </summary>
	public class Wabbitemu : IDisposable {

		private IWabbitemu ComObject;
		private uint Slot = 0;

		#region Types

		/// <summary>
		/// Provides access to the emulated calculator's memory.
		/// </summary>
		public class MemoryAccessor {
			private Wabbitemu Emulator;
			internal MemoryAccessor(Wabbitemu emulator) { this.Emulator = emulator; }

			/// <summary>
			/// Reads a byte of memory from the specified address.
			/// </summary>
			/// <param name="address">The address to read from.</param>
			/// <returns>The value read from that address.</returns>
			public byte Read(ushort address) {
				byte value;
				Marshal.ThrowExceptionForHR(this.Emulator.ComObject.ReadMem(this.Emulator.Slot, address, out value));
				return value;
			}

			/// <summary>
			/// Writes a byte of memory to the specified address.
			/// </summary>
			/// <param name="address">The address to write to.</param>
			/// <param name="value">The value to write at that address.</param>
			public void Write(ushort address, byte value) {
				Marshal.ThrowExceptionForHR(this.Emulator.ComObject.WriteMem(this.Emulator.Slot, address, value));
			}

			/// <summary>
			/// Gets or sets the contents of memory at a particular address.
			/// </summary>
			/// <param name="address">The address to read from or write to.</param>
			/// <returns>The contents of memory at the specified address.</returns>
			public byte this[ushort address] {
				get { return this.Read(address); }
				set { this.Write(address, value); }
			}
		}

		[Flags()]
		public enum MemoryTypes {
			None = 0,
			Ram = 1,
			Archive = 2,
			Both = Ram | Archive,
		}

		#endregion

		#region Constructor

		/// <summary>
		/// Creates an instance of the <see cref="Wabbitemu"/> class.
		/// </summary>
		/// <remarks>Wabbitemu must be running before an instance of this class is created.</remarks>
		public Wabbitemu() {
			var WabbitClassGuid = new Guid("8cc953bc-a879-492b-ad22-a2f4dfcd0e19");
			object o = null;
			try {
				o = Activator.CreateInstance(Type.GetTypeFromCLSID(WabbitClassGuid)); ;
			} catch (COMException ex) {
				if (ex.ErrorCode == unchecked((int)0x80040154)) { // "Class not registered".
					throw new InvalidOperationException("Wabbitemu is not running.");
				}
				throw;
			}
			try {
				this.ComObject = (IWabbitemu)o;
			} catch (InvalidCastException ex) {
				if (Marshal.GetHRForException(ex) == unchecked((int)0x80004002)) { // "No such interface".
					throw new InvalidOperationException("IWabbitemu.dll has not been registered.");
				}
				throw;
			}
		}

		~Wabbitemu() {
			this.Dispose();
		}

		#endregion

		#region Public Properties

		/// <summary>
		/// Gets the state of the calculator's Z80 CPU.
		/// </summary>
		public Z80State State {
			get {
				var S = new Z80State();
				Marshal.ThrowExceptionForHR(this.ComObject.GetState(this.Slot, out S));
				return S;
			}
			set {
				Marshal.ThrowExceptionForHR(this.ComObject.SetState(this.Slot, value));
			}
		}

		public MemoryAccessor Memory {
			get { return new MemoryAccessor(this); }
		}

		#endregion

		#region Public Methods

		/// <summary>
		/// Sends a key press message to Wabbitemu.
		/// </summary>
		/// <param name="key">The key to press.</param>
		public void KeyPress(Keys key) {
			Marshal.ThrowExceptionForHR(this.ComObject.KeyPress(this.Slot, key));
		}

		/// <summary>
		/// Sends a key release message to Wabbitemu.
		/// </summary>
		/// <param name="key">The key to release.</param>
		public void KeyRelease(Keys key) {
			Marshal.ThrowExceptionForHR(this.ComObject.KeyRelease(this.Slot, key));
		}

		/// <summary>
		/// Gets the screen data as a 128*64 array of bytes.
		/// </summary>
		public byte[] GetScreenData() {
			var Result = new byte[128 * 64];
			var ResultHandle = GCHandle.Alloc(Result, GCHandleType.Pinned);
			try {
				Marshal.ThrowExceptionForHR(this.ComObject.DrawScreen(this.Slot, ResultHandle.AddrOfPinnedObject()));
			} finally {
				ResultHandle.Free();
			}
			return Result;
		}

		#endregion
		
		#region IDisposable Members

		public void Dispose() {
			if (this.ComObject != null) {
				Marshal.FinalReleaseComObject(this.ComObject);
				this.ComObject = null;
			}
		}

		#endregion

	}
}
